"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Character = void 0;
const Constants_1 = require("./Constants");
const Observer_1 = require("./Observer");
const Pathfinder_1 = require("./Pathfinder");
const Tools_1 = require("./Tools");
class Character extends Observer_1.Observer {
    constructor(userID, userAuth, characterID, g, serverData) {
        super(serverData, g);
        this.timeouts = new Map();
        this.achievements = new Map();
        this.bank = { gold: 0 };
        this.chests = new Map();
        this.nextSkill = new Map();
        this.ready = false;
        this.apiercing = 0;
        this.blast = 0;
        this.armor = 0;
        this.attack = 0;
        this.c = {};
        this.gold = 0;
        this.hp = 0;
        this.level = 1;
        this.max_hp = 1;
        this.max_mp = 1;
        this.moving = false;
        this.mp = 1;
        this.range = 1;
        this.resistance = 0;
        this.rip = true;
        this.rpiercing = 0;
        this.s = {};
        this.speed = 1;
        this.tp = false;
        this.lastSmartMove = Date.now();
        this.owner = userID;
        this.userAuth = userAuth;
        this.characterID = characterID;
    }
    async updateLoop() {
        if (!this.socket || this.socket.disconnected || !this.ready) {
            this.timeouts.set("updateLoop", setTimeout(async () => { this.updateLoop(); }, Constants_1.Constants.UPDATE_POSITIONS_EVERY_MS));
            return;
        }
        if (this.lastPositionUpdate === undefined) {
            this.updatePositions();
            this.timeouts.set("updateLoop", setTimeout(async () => { this.updateLoop(); }, Constants_1.Constants.UPDATE_POSITIONS_EVERY_MS));
            return;
        }
        if (this.lastAllEntities !== undefined && Date.now() - this.lastAllEntities > Constants_1.Constants.STALE_MONSTER_MS) {
            await this.requestEntitiesData().catch(() => { });
        }
        const msSinceLastUpdate = Date.now() - this.lastPositionUpdate;
        if (msSinceLastUpdate > Constants_1.Constants.UPDATE_POSITIONS_EVERY_MS) {
            this.updatePositions();
            this.timeouts.set("updateLoop", setTimeout(async () => { this.updateLoop(); }, Constants_1.Constants.UPDATE_POSITIONS_EVERY_MS));
            return;
        }
        else {
            this.timeouts.set("updateLoop", setTimeout(async () => { this.updateLoop(); }, Constants_1.Constants.UPDATE_POSITIONS_EVERY_MS - msSinceLastUpdate));
            return;
        }
    }
    parseCharacter(data) {
        this.updatePositions();
        for (const datum in data) {
            if (datum == "hitchhikers") {
                for (const [event, datum] of data.hitchhikers) {
                    if (event == "game_response") {
                        this.parseGameResponse(datum);
                    }
                    else if (event == "eval") {
                        this.parseEval(datum);
                    }
                }
            }
            else if (datum == "entities") {
                this.parseEntities(data[datum]);
            }
            else if (datum == "owner") {
            }
            else if (datum == "tp") {
            }
            else if (datum == "user") {
                this.bank = data.user;
                for (const datum in this.bank) {
                    if (Array.isArray(this.bank[datum])) {
                        this.bank[datum].length = Constants_1.Constants.BANK_PACK_SIZE;
                    }
                }
            }
            else {
                this[datum] = data[datum];
            }
        }
        if (!this.party)
            this.partyData = undefined;
        if (!this.damage_type && this.ctype)
            this.damage_type = this.G.classes[this.ctype].damage_type;
    }
    async parseEntities(data) {
        for (let i = 0; i < data.players.length; i++) {
            const player = data.players[i];
            if (player.id == this.id) {
                this.parseCharacter(player);
                data.players.splice(i, 1);
                break;
            }
        }
        super.parseEntities(data);
    }
    parseEval(data) {
        const skillReg1 = /skill_timeout\s*\(\s*['"](.+?)['"]\s*,?\s*(\d+\.?\d+?)?\s*\)/.exec(data.code);
        if (skillReg1) {
            const skill = skillReg1[1];
            let cooldown;
            if (skillReg1[2]) {
                cooldown = Number.parseFloat(skillReg1[2]);
            }
            else if (this.G.skills[skill].cooldown) {
                cooldown = this.G.skills[skill].cooldown;
            }
            this.setNextSkill(skill, new Date(Date.now() + Math.ceil(cooldown)));
            return;
        }
        const potReg = /pot_timeout\s*\(\s*(\d+\.?\d+?)\s*\)/.exec(data.code);
        if (potReg) {
            const cooldown = Number.parseFloat(potReg[1]);
            this.setNextSkill("use_hp", new Date(Date.now() + Math.ceil(cooldown)));
            this.setNextSkill("use_mp", new Date(Date.now() + Math.ceil(cooldown)));
            return;
        }
    }
    parseGameResponse(data) {
        if (typeof (data) == "object") {
            if (data.response == "cooldown") {
                const skill = data.skill;
                if (skill) {
                    const cooldown = data.ms;
                    this.setNextSkill(skill, new Date(Date.now() + Math.ceil(cooldown)));
                }
            }
            else if (data.response == "ex_condition") {
                delete this.s[data.name];
            }
            else if (data.response == "skill_success") {
                const cooldown = this.G.skills[data.name].cooldown;
                if (cooldown) {
                    this.setNextSkill(data.name, new Date(Date.now() + cooldown));
                }
            }
            else {
                console.debug("Game Response Data -----");
                console.debug(data);
            }
        }
        else if (typeof (data) == "string") {
            if (data == "resolve_skill") {
            }
            else {
                console.debug(`Game Response: ${data}`);
            }
        }
    }
    parseNewMap(data) {
        this.going_x = data.x;
        this.going_y = data.y;
        this.in = data.in;
        this.m = data.m;
        this.moving = false;
        super.parseNewMap(data);
    }
    setNextSkill(skill, next) {
        this.nextSkill.set(skill, next);
    }
    updatePositions() {
        if (this.lastPositionUpdate) {
            const msSinceLastUpdate = Date.now() - this.lastPositionUpdate;
            if (msSinceLastUpdate == 0)
                return;
            if (this.moving) {
                const distanceTravelled = this.speed * msSinceLastUpdate / 1000;
                const angle = Math.atan2(this.going_y - this.y, this.going_x - this.x);
                const distanceToGoal = Tools_1.Tools.distance({ x: this.x, y: this.y }, { x: this.going_x, y: this.going_y });
                if (distanceTravelled > distanceToGoal) {
                    this.moving = false;
                    this.x = this.going_x;
                    this.y = this.going_y;
                }
                else {
                    this.x = this.x + Math.cos(angle) * distanceTravelled;
                    this.y = this.y + Math.sin(angle) * distanceTravelled;
                }
            }
            for (const condition in this.s) {
                const newCooldown = this.s[condition].ms - msSinceLastUpdate;
                if (newCooldown <= 0)
                    delete this.s[condition];
                else
                    this.s[condition].ms = newCooldown;
            }
        }
        super.updatePositions();
    }
    async connect() {
        try {
            await super.connect(false, false);
        }
        catch (e) {
            return Promise.reject(e);
        }
        this.socket.on("disconnect", () => {
            this.ready = false;
        });
        this.socket.on("disconnect_reason", () => {
            this.ready = false;
        });
        this.socket.on("start", (data) => {
            this.going_x = data.x;
            this.going_y = data.y;
            this.moving = false;
            this.damage_type = this.G.classes[data.ctype].damage_type;
            this.parseCharacter(data);
            if (data.entities)
                this.parseEntities(data.entities);
            this.S = data.s_info;
            this.ready = true;
        });
        this.socket.on("achievement_progress", (data) => {
            this.achievements.set(data.name, data);
        });
        this.socket.on("chest_opened", (data) => {
            this.chests.delete(data.id);
        });
        this.socket.on("drop", (data) => {
            this.chests.set(data.id, data);
        });
        this.socket.on("eval", (data) => {
            this.parseEval(data);
        });
        this.socket.on("game_error", (data) => {
            if (typeof data == "string") {
                console.error(`Game Error: ${data}`);
            }
            else {
                console.error("Game Error ----------");
                console.error(data);
            }
        });
        this.socket.on("game_response", (data) => {
            this.parseGameResponse(data);
        });
        this.socket.on("party_update", (data) => {
            this.partyData = data;
        });
        this.socket.on("player", (data) => {
            this.parseCharacter(data);
        });
        this.socket.on("q_data", (data) => {
            if (data.q.upgrade)
                this.q.upgrade = data.q.upgrade;
            if (data.q.compound)
                this.q.compound = data.q.compound;
        });
        this.socket.on("upgrade", (data) => {
            if (data.type == "compound" && this.q.compound)
                delete this.q.compound;
            else if (data.type == "upgrade" && this.q.upgrade)
                delete this.q.upgrade;
        });
        this.socket.on("welcome", (data) => {
            this.server = data;
            this.socket.emit("loaded", {
                height: 1080,
                scale: 2,
                success: 1,
                width: 1920
            });
            this.socket.emit("auth", {
                auth: this.userAuth,
                character: this.characterID,
                height: 1080,
                no_graphics: "True",
                no_html: "1",
                passphrase: "",
                scale: 2,
                user: this.owner,
                width: 1920
            });
        });
        const connected = new Promise((resolve, reject) => {
            const failCheck = (data) => {
                if (typeof data == "string") {
                    this.socket.removeListener("start", startCheck);
                    this.socket.removeListener("disconnect_reason", failCheck2);
                    reject(`Failed to connect: ${data}`);
                }
                else {
                    this.socket.removeListener("start", startCheck);
                    this.socket.removeListener("disconnect_reason", failCheck2);
                    reject(`Failed to connect: ${data.message}`);
                }
            };
            const failCheck2 = (data) => {
                this.socket.removeListener("start", startCheck);
                this.socket.removeListener("game_error", failCheck);
                reject(`Failed to connect: ${data}`);
            };
            const startCheck = () => {
                this.socket.removeListener("game_error", failCheck);
                this.socket.removeListener("disconnect_reason", failCheck2);
                this.updateLoop();
                resolve();
            };
            setTimeout(() => {
                this.socket.removeListener("start", startCheck);
                this.socket.removeListener("game_error", failCheck);
                this.socket.removeListener("disconnect_reason", failCheck2);
                reject(`Failed to start within ${Constants_1.Constants.CONNECT_TIMEOUT_MS / 1000}s.`);
            }, Constants_1.Constants.CONNECT_TIMEOUT_MS);
            this.socket.once("start", startCheck);
            this.socket.once("game_error", failCheck);
            this.socket.once("disconnect_reason", failCheck2);
        });
        this.socket.open();
        return connected;
    }
    async disconnect() {
        console.warn("Disconnecting!");
        if (this.socket) {
            this.socket.disconnect();
            this.socket.removeAllListeners();
        }
        this.ready = false;
        for (const [, timer] of this.timeouts)
            clearTimeout(timer);
    }
    async requestEntitiesData() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [requestEntitiesData].");
        return new Promise((resolve, reject) => {
            const checkEntitiesEvent = (data) => {
                if (data.type == "all") {
                    this.socket.removeListener("entities", checkEntitiesEvent);
                    resolve(data);
                }
            };
            setTimeout(() => {
                this.socket.removeListener("entities", checkEntitiesEvent);
                reject(`requestEntitiesData timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("entities", checkEntitiesEvent);
            this.socket.emit("send_updates", {});
        });
    }
    async requestPlayerData() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [requestPlayerData].");
        return new Promise((resolve, reject) => {
            const checkPlayerEvent = (data) => {
                if (data.s.typing) {
                    this.socket.removeListener("player", checkPlayerEvent);
                    resolve(data);
                }
            };
            setTimeout(() => {
                this.socket.removeListener("player", checkPlayerEvent);
                reject(`requestPlayerData timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("player", checkPlayerEvent);
            this.socket.emit("property", { typing: true });
        });
    }
    acceptMagiport(name) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [acceptMagiport].");
        const acceptedMagiport = new Promise((resolve, reject) => {
            const magiportCheck = (data) => {
                if (data.effect == "magiport") {
                    this.socket.removeListener("new_map", magiportCheck);
                    resolve({ map: data.name, x: data.x, y: data.y });
                }
            };
            setTimeout(() => {
                this.socket.removeListener("new_map", magiportCheck);
                reject(`acceptMagiport timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("new_map", magiportCheck);
        });
        this.socket.emit("magiport", { name: name });
        return acceptedMagiport;
    }
    acceptPartyInvite(id) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [acceptPartyInvite].");
        const acceptedInvite = new Promise((resolve, reject) => {
            const partyCheck = (data) => {
                if (data.list && data.list.includes(this.id)
                    && data.list.includes(id)) {
                    this.socket.removeListener("party_update", partyCheck);
                    this.socket.removeListener("game_log", unableCheck);
                    resolve(data);
                }
            };
            const unableCheck = (data) => {
                const notFound = RegExp("^.+? is not found$");
                if (data == "Invitation expired") {
                    this.socket.removeListener("party_update", partyCheck);
                    this.socket.removeListener("game_log", unableCheck);
                    reject(data);
                }
                else if (notFound.test(data)) {
                    this.socket.removeListener("party_update", partyCheck);
                    this.socket.removeListener("game_log", unableCheck);
                    reject(data);
                }
                else if (data == "Already partying") {
                    if (this.partyData.list.includes(this.id) && this.partyData.list.includes(id)) {
                        this.socket.removeListener("party_update", partyCheck);
                        this.socket.removeListener("game_log", unableCheck);
                        resolve(this.partyData);
                    }
                    else {
                        this.socket.removeListener("party_update", partyCheck);
                        this.socket.removeListener("game_log", unableCheck);
                        reject(data);
                    }
                }
            };
            setTimeout(() => {
                this.socket.removeListener("party_update", partyCheck);
                this.socket.removeListener("game_log", unableCheck);
                reject(`acceptPartyInvite timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("party_update", partyCheck);
            this.socket.on("game_log", unableCheck);
        });
        this.socket.emit("party", { event: "accept", name: id });
        return acceptedInvite;
    }
    acceptPartyRequest(id) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [acceptPartyRequest].");
        const acceptedRequest = new Promise((resolve, reject) => {
            const partyCheck = (data) => {
                if (data.list.includes(this.id)
                    && data.list.includes(id)) {
                    this.socket.removeListener("party_update", partyCheck);
                    resolve(data);
                }
            };
            setTimeout(() => {
                this.socket.removeListener("party_update", partyCheck);
                reject(`acceptPartyRequest timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("party_update", partyCheck);
        });
        this.socket.emit("party", { event: "raccept", name: id });
        return acceptedRequest;
    }
    basicAttack(id) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [basicAttack].");
        const attackStarted = new Promise((resolve, reject) => {
            const deathCheck = (data) => {
                if (data.id == id) {
                    this.socket.removeListener("action", attackCheck);
                    this.socket.removeListener("game_response", failCheck);
                    this.socket.removeListener("death", deathCheck);
                    reject(`Entity ${id} not found`);
                }
            };
            const failCheck = (data) => {
                if (typeof data == "object") {
                    if (data.response == "disabled") {
                        this.socket.removeListener("action", attackCheck);
                        this.socket.removeListener("game_response", failCheck);
                        this.socket.removeListener("death", deathCheck);
                        reject(`Attack on ${id} failed (disabled).`);
                    }
                    else if (data.response == "attack_failed" && data.id == id) {
                        this.socket.removeListener("action", attackCheck);
                        this.socket.removeListener("game_response", failCheck);
                        this.socket.removeListener("death", deathCheck);
                        reject(`Attack on ${id} failed.`);
                    }
                    else if (data.response == "too_far" && data.id == id) {
                        this.socket.removeListener("action", attackCheck);
                        this.socket.removeListener("game_response", failCheck);
                        this.socket.removeListener("death", deathCheck);
                        reject(`${id} is too far away to attack (dist: ${data.dist}).`);
                    }
                    else if (data.response == "cooldown" && data.id == id) {
                        this.socket.removeListener("action", attackCheck);
                        this.socket.removeListener("game_response", failCheck);
                        this.socket.removeListener("death", deathCheck);
                        reject(`Attack on ${id} failed due to cooldown (ms: ${data.ms}).`);
                    }
                    else if (data.response == "no_mp" && data.place == "attack") {
                        this.socket.removeListener("action", attackCheck);
                        this.socket.removeListener("game_response", failCheck);
                        this.socket.removeListener("death", deathCheck);
                        reject(`Attack on ${id} failed due to insufficient MP.`);
                    }
                }
            };
            const attackCheck = (data) => {
                if (data.attacker == this.id && data.target == id && data.type == "attack") {
                    this.socket.removeListener("action", attackCheck);
                    this.socket.removeListener("game_response", failCheck);
                    this.socket.removeListener("death", deathCheck);
                    resolve(data.pid);
                }
            };
            setTimeout(() => {
                this.socket.removeListener("action", attackCheck);
                this.socket.removeListener("game_response", failCheck);
                this.socket.removeListener("death", deathCheck);
                reject(`attack timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("action", attackCheck);
            this.socket.on("game_response", failCheck);
            this.socket.on("death", deathCheck);
        });
        this.socket.emit("attack", { id: id });
        return attackStarted;
    }
    buy(itemName, quantity = 1) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [buy].");
        if (this.gold < this.G.items[itemName].g)
            return Promise.reject(`Insufficient gold. We only have ${this.gold}, but the item costs ${this.G.items[itemName].g}`);
        const itemReceived = new Promise((resolve, reject) => {
            const buyCheck1 = (data) => {
                if (!data.hitchhikers)
                    return;
                for (const hitchhiker of data.hitchhikers) {
                    if (hitchhiker[0] == "game_response") {
                        const data = hitchhiker[1];
                        if (typeof data == "object"
                            && data.response == "buy_success"
                            && data.name == itemName
                            && data.q == quantity) {
                            this.socket.removeListener("player", buyCheck1);
                            this.socket.removeListener("game_response", buyCheck2);
                            resolve(data.num);
                        }
                    }
                }
            };
            const buyCheck2 = (data) => {
                if (data == "buy_cant_npc") {
                    this.socket.removeListener("player", buyCheck1);
                    this.socket.removeListener("game_response", buyCheck2);
                    reject(`Cannot buy ${quantity} ${itemName}(s) from an NPC`);
                }
                else if (data == "buy_cant_space") {
                    this.socket.removeListener("player", buyCheck1);
                    this.socket.removeListener("game_response", buyCheck2);
                    reject(`Not enough inventory space to buy ${quantity} ${itemName}(s)`);
                }
                else if (data == "buy_cost") {
                    this.socket.removeListener("player", buyCheck1);
                    this.socket.removeListener("game_response", buyCheck2);
                    reject(`Not enough gold to buy ${quantity} ${itemName}(s)`);
                }
            };
            setTimeout(() => {
                this.socket.removeListener("player", buyCheck1);
                this.socket.removeListener("game_response", buyCheck2);
                reject(`buy timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("player", buyCheck1);
            this.socket.on("game_response", buyCheck2);
        });
        if (this.G.items[itemName].s) {
            this.socket.emit("buy", { name: itemName, quantity: quantity });
        }
        else {
            this.socket.emit("buy", { name: itemName });
        }
        return itemReceived;
    }
    buyFromMerchant(id, slot, rid, quantity = 1) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [buyFromMerchant].");
        if (quantity <= 0)
            return Promise.reject(`We can not buy a quantity of ${quantity}.`);
        const merchant = this.players.get(id);
        if (!merchant)
            return Promise.reject(`We can not see ${id} nearby.`);
        if (Tools_1.Tools.distance(this, merchant) > Constants_1.Constants.NPC_INTERACTION_DISTANCE)
            return Promise.reject(`We are too far away from ${id} to buy from.`);
        const item = merchant.slots[slot];
        if (!item)
            return Promise.reject(`We could not find an item in slot ${slot} on ${id}.`);
        if (item.b)
            return Promise.reject("The item is not for sale, this merchant is *buying* that item.");
        if (item.rid !== rid)
            return Promise.reject(`The RIDs do not match (item: ${item.rid}, supplied: ${rid})`);
        if (!merchant.slots[slot].q && quantity != 1) {
            console.warn("We are only going to buy 1, as there is only 1 available.");
            quantity = 1;
        }
        else if (merchant.slots[slot].q && quantity > merchant.slots[slot].q) {
            console.warn(`We can't buy ${quantity}, we can only buy ${merchant.slots[slot].q}, so we're doing that.`);
            quantity = merchant.slots[slot].q;
        }
        if (this.gold < merchant.slots[slot].price * quantity) {
            if (this.gold < merchant.slots[slot].price)
                return Promise.reject(`We don't have enough gold. It costs ${merchant.slots[slot].price}, but we only have ${this.gold}`);
            const buyableQuantity = Math.floor(this.gold / merchant.slots[slot].price);
            console.warn(`We don't have enough gold to buy ${quantity}, we can only buy ${buyableQuantity}, so we're doing that.`);
            quantity = buyableQuantity;
        }
        this.socket.emit("trade_buy", { id: id, q: quantity.toString(), rid: rid, slot: slot });
    }
    buyFromPonty(item) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [buyFromPonty].");
        if (!item.rid)
            return Promise.reject("This item does not have an 'rid'.");
        const price = this.G.items[item.name].g * Constants_1.Constants.PONTY_MARKUP * (item.q ? item.q : 1);
        if (price > this.gold)
            return Promise.reject("We don't have enough gold to buy this.");
        const numBefore = this.countItem(item.name, this.items);
        const bought = new Promise((resolve, reject) => {
            const failCheck = (message) => {
                if (message == "Item gone") {
                    this.socket.removeListener("game_log", failCheck);
                    this.socket.removeListener("player", successCheck);
                    reject(`${item.name} is no longer available from Ponty.`);
                }
            };
            const successCheck = (data) => {
                const numNow = this.countItem(item.name, data.items);
                if ((item.q && numNow == numBefore + item.q)
                    || (numNow == numBefore + 1)) {
                    this.socket.removeListener("game_log", failCheck);
                    this.socket.removeListener("player", successCheck);
                    resolve();
                }
            };
            setTimeout(() => {
                this.socket.removeListener("game_log", failCheck);
                this.socket.removeListener("player", successCheck);
                reject(`buyFromPonty timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("game_log", failCheck);
            this.socket.on("player", successCheck);
        });
        this.socket.emit("sbuy", { rid: item.rid });
        return bought;
    }
    calculateDamageRange(defender, skill = "attack") {
        if (defender["1hp"])
            return [1, 1];
        let baseDamage = this.attack;
        if (this.G.skills[skill].damage)
            baseDamage = this.G.skills[skill].damage;
        if (defender.s.cursed)
            baseDamage *= 1.2;
        if (defender.s.marked)
            baseDamage *= 1.1;
        if (this.ctype == "priest")
            baseDamage *= 0.4;
        let additonalApiercing = 0;
        if (this.G.skills[skill].apiercing)
            additonalApiercing = this.G.skills[skill].apiercing;
        if (this.damage_type == "physical")
            baseDamage *= Tools_1.Tools.damage_multiplier(defender.armor - this.apiercing - additonalApiercing);
        else if (this.damage_type == "magical")
            baseDamage *= Tools_1.Tools.damage_multiplier(defender.resistance - this.rpiercing);
        if (this.G.skills[skill].damage_multiplier)
            baseDamage *= this.G.skills[skill].damage_multiplier;
        if (this.crit) {
            if (this.crit >= 100) {
                return [baseDamage * 0.9 * (2 + (this.critdamage / 100)), baseDamage * 1.1 * (2 + (this.critdamage / 100))];
            }
            else {
                return [baseDamage * 0.9, baseDamage * 1.1 * (2 + (this.critdamage / 100))];
            }
        }
        else {
            return [baseDamage * 0.9, baseDamage * 1.1];
        }
    }
    calculateItemCost(item) {
        const gInfo = this.G.items[item.name];
        let cost = gInfo.g;
        if (gInfo.compound) {
            for (let i = 0; i < item.level; i++) {
                cost *= 3;
                let scrollLevel = 0;
                for (const grade of gInfo.grades) {
                    if (i + 1 < grade) {
                        const scrollInfo = this.G.items[`cscroll${scrollLevel}`];
                        cost += scrollInfo.g;
                        break;
                    }
                    scrollLevel++;
                }
            }
        }
        else if (gInfo.upgrade) {
            for (let i = 0; i < item.level; i++) {
                let scrollLevel = 0;
                for (const grade of gInfo.grades) {
                    if (i + 1 < grade) {
                        const scrollInfo = this.G.items[`scroll${scrollLevel}`];
                        cost += scrollInfo.g;
                        break;
                    }
                    scrollLevel++;
                }
            }
        }
        if (item.gift)
            cost -= (gInfo.g - 1);
        return cost;
    }
    calculateItemGrade(item) {
        const gInfo = this.G.items[item.name];
        if (!gInfo.grades)
            return;
        let grade = 0;
        for (const level of gInfo.grades) {
            if (item.level < level)
                break;
            grade++;
        }
        return grade;
    }
    canBuy(item) {
        if (this.isFull())
            return false;
        const gInfo = this.G.items[item];
        if (this.gold < gInfo.g)
            return false;
        const computerAvailable = this.locateItem("computer") !== undefined;
        let buyable = false;
        let close = false;
        for (const map in this.G.maps) {
            if (buyable == true)
                break;
            if (!computerAvailable && map !== this.map)
                continue;
            if (this.G.maps[map].ignore)
                continue;
            for (const npc of this.G.maps[map].npcs) {
                if (buyable == true)
                    break;
                if (this.G.npcs[npc.id].items === undefined)
                    continue;
                for (const i of this.G.npcs[npc.id].items) {
                    if (i == item) {
                        buyable = true;
                        if (Tools_1.Tools.distance(this, { map: map, x: npc.position[0], y: npc.position[1] }) < Constants_1.Constants.NPC_INTERACTION_DISTANCE)
                            close = true;
                        break;
                    }
                }
            }
        }
        if (!buyable)
            return false;
        if (computerAvailable || close)
            return true;
        return false;
    }
    canCraft(itemToCraft, options) {
        const gCraft = this.G.craft[itemToCraft];
        if (!gCraft)
            return false;
        if (gCraft.cost > this.gold)
            return false;
        for (const [requiredQuantity, requiredItem, requiredItemLevel] of gCraft.items) {
            if (!this.hasItem(requiredItem, this.items, { level: requiredItemLevel, quantityGreaterThan: requiredQuantity - 1 }))
                return false;
        }
        if (this.G.maps[this.map].mount)
            return false;
        if (!this.hasItem("computer") && !options?.ignoreLocation) {
            const craftableLocation = this.locateCraftNPC(itemToCraft);
            if (Tools_1.Tools.distance(this, craftableLocation) > Constants_1.Constants.NPC_INTERACTION_DISTANCE)
                return false;
        }
        return true;
    }
    canExchange(itemToExchange, options) {
        const gItem = this.G.items[itemToExchange];
        if (gItem.e !== undefined && this.countItem(itemToExchange) < gItem.e)
            return false;
        if (!this.hasItem("computer") && !options?.ignoreLocation) {
            const exchangeableLocation = this.locateExchangeNPC(itemToExchange);
            if (Tools_1.Tools.distance(this, exchangeableLocation) > Constants_1.Constants.NPC_INTERACTION_DISTANCE)
                return false;
        }
        return true;
    }
    canKillInOneShot(entity, skill = "attack") {
        const gInfo = this.G.monsters[entity.type];
        if (gInfo.lifesteal !== undefined)
            return false;
        if (gInfo.abilities && gInfo.abilities.self_healing)
            return false;
        if (entity.avoidance)
            return false;
        if (this.damage_type == "magical" && entity.reflection)
            return false;
        if (this.damage_type == "physical" && entity.evasion)
            return false;
        if (entity["1hp"]) {
            return entity.hp == 1;
        }
        let minimumDamage = this.calculateDamageRange(entity)[0];
        if (this.G.skills[skill].damage_multiplier)
            minimumDamage *= this.G.skills[skill].damage_multiplier;
        return minimumDamage > entity.hp;
    }
    canSell() {
        if (this.hasItem("computer"))
            return true;
        for (const npc of this.G.maps[this.map].npcs) {
            const gNPC = this.G.npcs[npc.id];
            if (!gNPC.items)
                continue;
            if (Tools_1.Tools.distance(this, { map: this.map, x: npc.position[0], y: npc.position[1] }) > Constants_1.Constants.NPC_INTERACTION_DISTANCE)
                continue;
            return true;
        }
        return false;
    }
    canUse(skill, options) {
        if (this.rip)
            return false;
        if (this.s.stoned)
            return false;
        if (this.isOnCooldown(skill) && !options?.ignoreCooldown)
            return false;
        const gInfoSkill = this.G.skills[skill];
        if (gInfoSkill.mp !== undefined && this.mp < gInfoSkill.mp)
            return false;
        if (skill == "attack" && this.mp < this.mp_cost)
            return false;
        if (gInfoSkill.level !== undefined && this.level < gInfoSkill.level)
            return false;
        if (gInfoSkill.wtype && !options?.ignoreEquipped) {
            if (!this.slots.mainhand)
                return false;
            const gInfoWeapon = this.G.items[this.slots.mainhand.name];
            if (typeof gInfoSkill.wtype == "object") {
                let isAcceptableWeapon = false;
                for (const wtype of gInfoSkill.wtype) {
                    if (gInfoWeapon.wtype == wtype) {
                        isAcceptableWeapon = true;
                        break;
                    }
                }
                if (!isAcceptableWeapon)
                    return false;
            }
            else {
                if (gInfoWeapon.wtype !== gInfoSkill.wtype)
                    return false;
            }
        }
        if (gInfoSkill.consume && !options?.ignoreEquipped) {
            if (!this.hasItem(gInfoSkill.consume))
                return false;
        }
        if (gInfoSkill.inventory && !options?.ignoreEquipped) {
            for (const item of gInfoSkill.inventory) {
                if (!this.hasItem(item))
                    return false;
            }
        }
        if (gInfoSkill.slot && !options?.ignoreEquipped) {
            let hasSlot = false;
            for (const [slot, item] of gInfoSkill.slot) {
                if (this.slots[slot] && this.slots[slot].name == item) {
                    hasSlot = true;
                    break;
                }
            }
            if (!hasSlot)
                return false;
        }
        if (gInfoSkill.class) {
            let compatibleClass = false;
            for (const c of gInfoSkill.class) {
                if (c == this.ctype) {
                    compatibleClass = true;
                    break;
                }
            }
            if (!compatibleClass)
                return false;
        }
        if (gInfoSkill.requirements) {
            for (const s in gInfoSkill.requirements) {
                const stat = s;
                if (this[stat] < gInfoSkill.requirements[stat])
                    return false;
            }
        }
        if (this.s.dampened) {
            if (skill == "blink")
                return false;
        }
        if (this.ctype == "merchant" && skill == "attack") {
            if (!this.slots.mainhand)
                return false;
            if (this.slots.mainhand.name !== "dartgun")
                return false;
            if (this.gold < 100)
                return false;
        }
        return true;
    }
    compound(item1Pos, item2Pos, item3Pos, cscrollPos, offeringPos) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [compound].");
        const item1Info = this.items[item1Pos];
        const item2Info = this.items[item2Pos];
        const item3Info = this.items[item3Pos];
        const cscrollInfo = this.items[cscrollPos];
        if (!item1Info)
            return Promise.reject(`There is no item in inventory slot ${item1Pos} (item1).`);
        if (!item2Info)
            return Promise.reject(`There is no item in inventory slot ${item2Pos} (item2).`);
        if (!item3Info)
            return Promise.reject(`There is no item in inventory slot ${item3Pos} (item3).`);
        if (!cscrollInfo)
            return Promise.reject(`There is no item in inventory slot ${cscrollPos} (cscroll).`);
        if (offeringPos !== undefined) {
            const offeringInfo = this.items[offeringPos];
            if (!offeringInfo)
                return Promise.reject(`There is no item in inventory slot ${offeringPos} (offering).`);
        }
        if (item1Info.name != item2Info.name || item1Info.name != item3Info.name)
            return Promise.reject("You can only combine 3 of the same items.");
        if (item1Info.level != item2Info.level || item1Info.level != item3Info.level)
            return Promise.reject("You can only combine 3 items of the same level.");
        const compoundComplete = new Promise((resolve, reject) => {
            const playerCheck = (data) => {
                if (!data.hitchhikers)
                    return;
                for (const [event, datum] of data.hitchhikers) {
                    if (event == "game_response" && datum.response == "compound_fail") {
                        this.socket.removeListener("game_response", gameResponseCheck);
                        this.socket.removeListener("player", playerCheck);
                        resolve(false);
                        return;
                    }
                    else if (event == "game_response" && datum.response == "compound_success") {
                        this.socket.removeListener("game_response", gameResponseCheck);
                        this.socket.removeListener("player", playerCheck);
                        resolve(true);
                        return;
                    }
                }
            };
            const gameResponseCheck = (data) => {
                if (typeof data == "object") {
                    if (data.response == "bank_restrictions" && data.place == "compound") {
                        this.socket.removeListener("game_response", gameResponseCheck);
                        this.socket.removeListener("player", playerCheck);
                        reject("You can't compound items in the bank.");
                    }
                }
                else if (typeof data == "string") {
                    if (data == "compound_no_item") {
                        this.socket.removeListener("game_response", gameResponseCheck);
                        this.socket.removeListener("player", playerCheck);
                        reject();
                    }
                }
            };
            setTimeout(() => {
                this.socket.removeListener("game_response", gameResponseCheck);
                this.socket.removeListener("player", playerCheck);
                reject("compound timeout (60000ms)");
            }, 60000);
            this.socket.on("game_response", gameResponseCheck);
            this.socket.on("player", playerCheck);
        });
        this.socket.emit("compound", {
            "clevel": item1Info.level,
            "items": [item1Pos, item2Pos, item3Pos],
            "offering_num": offeringPos,
            "scroll_num": cscrollPos
        });
        return compoundComplete;
    }
    craft(item) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [craft].");
        const gInfo = this.G.craft[item];
        if (!gInfo)
            return Promise.reject(`Can not find a recipe for ${item}.`);
        if (gInfo.cost > this.gold)
            return Promise.reject(`We don't have enough gold to craft ${item}.`);
        const itemPositions = [];
        for (let i = 0; i < gInfo.items.length; i++) {
            const requiredQuantity = gInfo.items[i][0];
            const requiredName = gInfo.items[i][1];
            const requiredLevel = gInfo.items[i][2];
            const searchArgs = {
                level: requiredLevel,
                quantityGreaterThan: requiredQuantity > 1 ? requiredQuantity : undefined,
            };
            const itemPos = this.locateItem(requiredName, this.items, searchArgs);
            if (itemPos == undefined)
                return Promise.reject(`We don't have ${requiredQuantity} ${requiredName} to craft ${item}.`);
            itemPositions.push([i, itemPos]);
        }
        const crafted = new Promise((resolve, reject) => {
            const successCheck = async (data) => {
                if (typeof data == "object") {
                    if (data.response == "craft" && data.name == item) {
                        this.socket.removeListener("game_response", successCheck);
                        resolve();
                    }
                }
            };
            setTimeout(() => {
                this.socket.removeListener("game_response", successCheck);
                reject(`craft timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
        });
        this.socket.emit("craft", { items: itemPositions });
        return crafted;
    }
    depositGold(gold) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [depositGold].");
        if (this.map !== "bank")
            return Promise.reject("We need to be in 'bank' to deposit gold.");
        if (gold <= 0)
            return Promise.reject("We can't deposit 0 or less gold");
        if (gold > this.gold) {
            gold = this.gold;
            console.warn(`We are only going to deposit ${gold} gold.`);
        }
        this.socket.emit("bank", { amount: gold, operation: "deposit" });
    }
    depositItem(inventoryPos, bankPack, bankSlot = -1) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [depositItem].");
        if (this.map !== "bank" && this.map !== "bank_b" && this.map !== "bank_u")
            return Promise.reject(`We're not in the bank (we're in '${this.map}')`);
        const item = this.items[inventoryPos];
        if (!item)
            return Promise.reject(`There is no item in inventory slot ${inventoryPos}.`);
        if (bankPack) {
            const bankPackNum = Number.parseInt(bankPack.substr(5, 2));
            if ((this.map == "bank" && bankPackNum < 0 && bankPackNum > 7)
                || (this.map == "bank_b" && bankPackNum < 8 && bankPackNum > 23)
                || (this.map == "bank_u" && bankPackNum < 24 && bankPackNum > 47)) {
                return Promise.reject(`We can not access ${bankPack} on ${this.map}.`);
            }
        }
        else {
            bankSlot = undefined;
            let packFrom;
            let packTo;
            if (this.map == "bank") {
                packFrom = 0;
                packTo = 7;
            }
            else if (this.map == "bank_b") {
                packFrom = 8;
                packTo = 23;
            }
            else if (this.map == "bank_u") {
                packFrom = 24;
                packTo = 47;
            }
            const numStackable = this.G.items[item.name].s;
            let emptyPack;
            let emptySlot;
            for (let packNum = packFrom; packNum <= packTo; packNum++) {
                const packName = `items${packNum}`;
                const pack = this.bank[packName];
                if (!pack)
                    continue;
                for (let slotNum = 0; slotNum < pack.length; slotNum++) {
                    const slot = pack[slotNum];
                    if (!slot) {
                        if (!numStackable) {
                            bankPack = packName;
                            bankSlot = slotNum;
                            break;
                        }
                        else if (!emptyPack && emptySlot == undefined) {
                            emptyPack = packName;
                            emptySlot = slotNum;
                        }
                    }
                    else if (numStackable && slot.name == item.name && (slot.q + item.q <= numStackable)) {
                        bankPack = packName;
                        bankSlot = -1;
                        break;
                    }
                }
                if (bankPack && bankSlot !== undefined)
                    break;
            }
            if (bankPack == undefined && bankSlot == undefined && emptyPack !== undefined && emptySlot !== undefined) {
                bankPack = emptyPack;
                bankSlot = emptySlot;
            }
            else if (bankPack === undefined && bankSlot === undefined && emptyPack === undefined && emptySlot === undefined) {
                return Promise.reject(`Bank is full. There is nowhere to place '${item.name}'.`);
            }
        }
        const bankItemCount = this.countItem(item.name, this.bank[bankPack]);
        const swapped = new Promise((resolve, reject) => {
            const checkDeposit = (data) => {
                if (!data.user) {
                    if (data.map !== "bank" && data.map !== "bank_b" && data.map !== "bank_u") {
                        this.socket.removeListener("player", checkDeposit);
                        return reject(`We're not in the bank (we're in '${data.map}')`);
                    }
                }
                else {
                    const newBankItemCount = this.countItem(item.name, data.user[bankPack]);
                    if ((item.q && newBankItemCount == (bankItemCount + item.q))
                        || (!item.q && newBankItemCount == (bankItemCount + 1))) {
                        this.socket.removeListener("player", checkDeposit);
                        return resolve();
                    }
                }
            };
            setTimeout(() => {
                this.socket.removeListener("player", checkDeposit);
                reject(`depositItem timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("player", checkDeposit);
        });
        this.socket.emit("bank", { inv: inventoryPos, operation: "swap", pack: bankPack, str: bankSlot });
        return swapped;
    }
    emote(emotionName) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [emote].");
        if (!this.emx[emotionName])
            return Promise.reject(`We don't have the emotion '${emotionName}'`);
        const emoted = new Promise((resolve, reject) => {
            const failCheck = (data) => {
                if (typeof data == "string") {
                    if (data == "emotion_cooldown") {
                        this.socket.removeListener("game_response", failCheck);
                        this.socket.removeListener("emotion", successCheck);
                        reject();
                    }
                    else if (data == "emotion_cant") {
                        this.socket.removeListener("game_response", failCheck);
                        this.socket.removeListener("emotion", successCheck);
                        reject();
                    }
                }
            };
            const successCheck = (data) => {
                if (data.name == emotionName && data.player == this.name) {
                    this.socket.removeListener("game_response", failCheck);
                    this.socket.removeListener("emotion", successCheck);
                    resolve();
                }
            };
            setTimeout(() => {
                this.socket.removeListener("game_response", failCheck);
                this.socket.removeListener("emotion", successCheck);
                reject(`emote timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("game_response", failCheck);
            this.socket.on("emotion", successCheck);
        });
        this.socket.emit("emotion", { name: emotionName });
        return emoted;
    }
    equip(inventoryPos, equipSlot) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [equip].");
        if (!this.items[inventoryPos])
            return Promise.reject(`No item in inventory slot ${inventoryPos}.`);
        const iInfo = this.items[inventoryPos];
        const equipFinished = new Promise((resolve, reject) => {
            const equipCheck = (data) => {
                if (equipSlot) {
                    const item = data.slots[equipSlot];
                    if (item
                        && item.name == iInfo.name
                        && item.level == iInfo.level
                        && item.p == iInfo.p) {
                        this.socket.removeListener("player", equipCheck);
                        this.socket.removeListener("disappearing_text", cantEquipCheck);
                        resolve();
                    }
                }
                else {
                    for (const slot in data.slots) {
                        const item = data.slots[slot];
                        if (item && item.name == iInfo.name) {
                            this.socket.removeListener("player", equipCheck);
                            this.socket.removeListener("disappearing_text", cantEquipCheck);
                            resolve();
                        }
                    }
                }
            };
            const cantEquipCheck = (data) => {
                if (data.id == this.id && data.message == "CAN'T EQUIP") {
                    this.socket.removeListener("player", equipCheck);
                    this.socket.removeListener("disappearing_text", cantEquipCheck);
                    reject(`Can't equip '${inventoryPos}' (${iInfo.name})`);
                }
            };
            setTimeout(() => {
                this.socket.removeListener("player", equipCheck);
                this.socket.removeListener("disappearing_text", cantEquipCheck);
                reject(`equip timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("player", equipCheck);
            this.socket.on("disappearing_text", cantEquipCheck);
        });
        this.socket.emit("equip", { num: inventoryPos, slot: equipSlot });
        return equipFinished;
    }
    exchange(inventoryPos) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [exchange].");
        if (!this.items[inventoryPos])
            return Promise.reject(`No item in inventory slot ${inventoryPos}.`);
        if (this.G.maps[this.map].mount)
            return Promise.reject("We can't exchange things in the bank.");
        let startedExchange = false;
        if (this.q.exchange)
            startedExchange = true;
        const exchangeFinished = new Promise((resolve, reject) => {
            const completeCheck = (data) => {
                if (!startedExchange && data.q.exchange?.len == data.q.exchange?.ms) {
                    startedExchange = true;
                    return;
                }
                if (startedExchange && !data.q.exchange) {
                    this.socket.removeListener("player", completeCheck);
                    this.socket.removeListener("game_response", bankCheck);
                    resolve();
                }
            };
            const bankCheck = (data) => {
                if (typeof data == "object" && data.response == "bank_restrictions" && data.place == "upgrade") {
                    this.socket.removeListener("player", completeCheck);
                    this.socket.removeListener("game_response", bankCheck);
                    reject("You can't exchange items in the bank.");
                }
                else if (typeof data == "string") {
                    if (data == "exchange_notenough") {
                        this.socket.removeListener("player", completeCheck);
                        this.socket.removeListener("game_response", bankCheck);
                        reject("We don't have enough items to exchange.");
                    }
                    else if (data == "exchange_existing") {
                        this.socket.removeListener("player", completeCheck);
                        this.socket.removeListener("game_response", bankCheck);
                        reject("We are already exchanging something.");
                    }
                }
            };
            setTimeout(() => {
                this.socket.removeListener("player", completeCheck);
                this.socket.removeListener("game_response", bankCheck);
                reject("exchange timeout (60000ms)");
            }, 60000);
            this.socket.on("game_response", bankCheck);
            this.socket.on("player", completeCheck);
        });
        this.socket.emit("exchange", { item_num: inventoryPos, q: this.items[inventoryPos]?.q });
        return exchangeFinished;
    }
    finishMonsterHuntQuest() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [finishMonsterHuntQuest].");
        if (!this.s.monsterhunt)
            return Promise.reject("We don't have a monster hunt to turn in.");
        if (this.s.monsterhunt.c > 0)
            return Promise.reject(`We still have to kill ${this.s.monsterhunt.c} ${this.s.monsterhunt.id}(s).`);
        let close = false;
        for (const npc of this.G.maps[this.map].npcs) {
            if (npc.id !== "monsterhunter")
                continue;
            if (Tools_1.Tools.distance(this, { x: npc.position[0], y: npc.position[1] }) > Constants_1.Constants.NPC_INTERACTION_DISTANCE)
                continue;
            close = true;
            break;
        }
        if (!close)
            return Promise.reject("We are too far away from the Monster Hunter NPC.");
        const questFinished = new Promise((resolve, reject) => {
            const successCheck = (data) => {
                if (!data.s || data.s.monsterhunt == undefined) {
                    this.socket.removeListener("player", successCheck);
                    resolve();
                }
            };
            setTimeout(() => {
                this.socket.removeListener("player", successCheck);
                reject(`getMonsterHuntQuest timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("player", successCheck);
        });
        this.socket.emit("monsterhunt");
        return questFinished;
    }
    getEntities(filters = {}) {
        const entities = [];
        for (const [, entity] of this.entities) {
            if (filters.targetingMe !== undefined) {
                if (filters.targetingMe) {
                    if (entity.target !== this.id)
                        continue;
                }
                else {
                    if (entity.target == this.id)
                        continue;
                }
            }
            if (filters.targetingPlayer && entity.target !== filters.targetingPlayer)
                continue;
            if (filters.level !== entity.level)
                continue;
            if (filters.levelGreaterThan <= entity.level)
                continue;
            if (filters.levelLessThan >= entity.level)
                continue;
            if (filters.type !== entity.type)
                continue;
            if (filters.typeList && !filters.typeList.includes(entity.type))
                continue;
            if (filters.withinRange !== undefined && Tools_1.Tools.distance(this, entity) > filters.withinRange)
                continue;
            if (filters.canWalkTo !== undefined) {
                const canWalkTo = Pathfinder_1.Pathfinder.canWalkPath(this, entity);
                if (filters.canWalkTo && !canWalkTo)
                    continue;
                if (!filters.canWalkTo && canWalkTo)
                    continue;
            }
            if (filters.couldGiveCredit) {
                const couldGiveCredit = entity.couldGiveCreditForKill(this);
                if (filters.couldGiveCredit && !couldGiveCredit)
                    continue;
                if (!filters.couldGiveCredit && couldGiveCredit)
                    continue;
            }
            if (filters.willBurnToDeath) {
                const willBurnToDeath = entity.willBurnToDeath();
                if (filters.willBurnToDeath && !willBurnToDeath)
                    continue;
                if (!filters.willDieToProjectiles && willBurnToDeath)
                    continue;
            }
            if (filters.willDieToProjectiles) {
                const willDieToProjectiles = entity.willDieToProjectiles(this.projectiles, this.players, this.entities);
                if (filters.willDieToProjectiles && !willDieToProjectiles)
                    continue;
                if (!filters.willDieToProjectiles && willDieToProjectiles)
                    continue;
            }
            entities.push(entity);
        }
        return entities;
    }
    async getMonsterHuntQuest() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [getMonsterHuntQuest].");
        if (this.s.monsterhunt && this.s.monsterhunt.c > 0)
            return Promise.reject(`We can't get a new monsterhunt. We have ${this.s.monsterhunt.ms}ms left to kill ${this.s.monsterhunt.c} ${this.s.monsterhunt.id}(s).`);
        if (this.ctype == "merchant")
            return Promise.reject("Merchants can't do Monster Hunts.");
        let close = false;
        for (const npc of this.G.maps[this.map].npcs) {
            if (npc.id !== "monsterhunter")
                continue;
            if (Tools_1.Tools.distance(this, { x: npc.position[0], y: npc.position[1] }) > Constants_1.Constants.NPC_INTERACTION_DISTANCE)
                continue;
            close = true;
            break;
        }
        if (!close)
            return Promise.reject("We are too far away from the Monster Hunter NPC.");
        if (this.s.monsterhunt && this.s.monsterhunt.c == 0) {
            console.warn("We are going to finish the current monster quest first.");
            await this.finishMonsterHuntQuest();
        }
        const questGot = new Promise((resolve, reject) => {
            const failCheck = (data) => {
                if (data == "ecu_get_closer") {
                    this.socket.removeListener("game_response", failCheck);
                    this.socket.removeListener("player", successCheck);
                    reject("Too far away from Monster Hunt NPC.");
                }
                else if (data == "monsterhunt_merchant") {
                    this.socket.removeListener("game_response", failCheck);
                    this.socket.removeListener("player", successCheck);
                    reject("Merchants can't do Monster Hunts.");
                }
            };
            const successCheck = (data) => {
                if (!data.hitchhikers)
                    return;
                for (const hitchhiker of data.hitchhikers) {
                    if (hitchhiker[0] == "game_response" && hitchhiker[1] == "monsterhunt_started") {
                        this.socket.removeListener("game_response", failCheck);
                        this.socket.removeListener("player", successCheck);
                        resolve();
                        return;
                    }
                }
            };
            setTimeout(() => {
                this.socket.removeListener("game_response", failCheck);
                this.socket.removeListener("player", successCheck);
                reject(`getMonsterHuntQuest timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("game_response", failCheck);
            this.socket.on("player", successCheck);
        });
        this.socket.emit("monsterhunt");
        return questGot;
    }
    getPlayers() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [getPlayers].");
        const playersData = new Promise((resolve, reject) => {
            const dataCheck = (data) => {
                resolve(data);
            };
            setTimeout(() => {
                this.socket.removeListener("players", dataCheck);
                reject(`getPlayers timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.once("players", dataCheck);
        });
        this.socket.emit("players");
        return playersData;
    }
    getPontyItems() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [getPontyItems].");
        const pontyItems = new Promise((resolve, reject) => {
            const distanceCheck = (data) => {
                if (data == "buy_get_closer") {
                    this.socket.removeListener("game_response", distanceCheck);
                    this.socket.removeListener("secondhands", secondhandsItems);
                    reject("Too far away from secondhands NPC.");
                }
            };
            const secondhandsItems = (data) => {
                this.socket.removeListener("game_response", distanceCheck);
                this.socket.removeListener("secondhands", secondhandsItems);
                resolve(data);
            };
            setTimeout(() => {
                this.socket.removeListener("game_response", distanceCheck);
                this.socket.removeListener("secondhands", secondhandsItems);
                reject(`getPontyItems timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("secondhands", secondhandsItems);
            this.socket.on("game_response", distanceCheck);
        });
        this.socket.emit("secondhands");
        return pontyItems;
    }
    getTrackerData() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [getTrackerData].");
        if (!this.hasItem("tracker"))
            return Promise.reject("We need a tracker to obtain tracker data.");
        const gotTrackerData = new Promise((resolve, reject) => {
            const gotCheck = (data) => {
                resolve(data);
            };
            setTimeout(() => {
                this.socket.removeListener("tracker", gotCheck);
                reject(`getTrackerData timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.once("tracker", gotCheck);
        });
        this.socket.emit("tracker");
        return gotTrackerData;
    }
    isFull() {
        return this.esize == 0;
    }
    isScared() {
        return this.fear > 0;
    }
    kickPartyMember(toKick) {
        if (!this.party)
            return Promise.resolve();
        if (!this.partyData.list.includes(toKick))
            return Promise.resolve();
        if (toKick == this.id)
            return this.leaveParty();
        if (this.partyData.list.indexOf(this.id) > this.partyData.list.indexOf(toKick))
            return Promise.reject(`We can't kick ${toKick}, they're higher on the party list.`);
        const kicked = new Promise((resolve, reject) => {
            const kickedCheck = (data) => {
                if (!data.list || !data.list.includes(toKick)) {
                    this.socket.removeListener("party_update", kickedCheck);
                    resolve();
                }
            };
            setTimeout(() => {
                this.socket.removeListener("party_update", kickedCheck);
                reject(`kickPartyMember timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("party_update", kickedCheck);
        });
        this.socket.emit("party", { event: "kick", name: toKick });
        return kicked;
    }
    leaveMap() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [leaveMap].");
        const leaveComplete = new Promise((resolve, reject) => {
            const leaveCheck = (data) => {
                if (data.name == "main") {
                    this.socket.removeListener("new_map", leaveCheck);
                    this.socket.removeListener("game_response", failCheck);
                    resolve();
                }
                else {
                    this.socket.removeListener("new_map", leaveCheck);
                    this.socket.removeListener("game_response", failCheck);
                    reject(`We are now in ${data.name}, but we should be in main`);
                }
            };
            const failCheck = (data) => {
                if (typeof data == "string") {
                    if (data == "cant_escape") {
                        this.socket.removeListener("new_map", leaveCheck);
                        this.socket.removeListener("game_response", failCheck);
                        reject(`Can't escape from current map ${this.map}`);
                    }
                }
            };
            setTimeout(() => {
                this.socket.removeListener("new_map", leaveCheck);
                this.socket.removeListener("game_response", failCheck);
                reject(`leaveMap timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.once("new_map", leaveCheck);
            this.socket.on("game_response", failCheck);
        });
        this.socket.emit("leave");
        return leaveComplete;
    }
    leaveParty() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [leaveParty].");
        this.socket.emit("party", { event: "leave" });
    }
    async move(x, y, options) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [move].");
        if (this.x == x && this.y == y)
            return Promise.resolve({ map: this.map, x: this.x, y: this.y });
        let to = { map: this.map, x: x, y: y };
        if (!options?.disableSafetyCheck) {
            to = Pathfinder_1.Pathfinder.getSafeWalkTo({ map: this.map, x: this.x, y: this.y }, { map: this.map, x, y });
            if (to.x !== x || to.y !== y) {
                console.warn(`move: We can't move to {x: ${x}, y: ${y}} safely. We will move to {x: ${to.x}, y: ${to.y}}.`);
            }
        }
        const moveFinished = new Promise((resolve, reject) => {
            let timeToFinishMove = 1 + Tools_1.Tools.distance(this, { x: to.x, y: to.y }) / this.speed;
            const checkPlayer = async (data) => {
                if (!data.moving || data.going_x != to.x || data.going_y != to.y) {
                    const newData = await this.requestPlayerData();
                    if (!newData.moving || newData.going_x != to.x || newData.going_y != to.y) {
                        clearTimeout(timeout);
                        this.socket.removeListener("player", checkPlayer);
                        reject(`move to ${to.x}, ${to.y} failed`);
                    }
                }
                else {
                    timeToFinishMove = Tools_1.Tools.distance(this, { x: data.x, y: data.y }) / data.speed;
                    clearTimeout(timeout);
                    timeout = setTimeout(checkPosition, timeToFinishMove);
                }
            };
            const checkPosition = () => {
                this.updatePositions();
                timeToFinishMove = 1 + Tools_1.Tools.distance(this, { x: to.x, y: to.y }) / this.speed;
                if (this.x == to.x && this.y == to.y) {
                    this.socket.removeListener("player", checkPlayer);
                    resolve({ map: this.map, x: x, y: y });
                }
                else if (this.moving && this.going_x == to.x && this.going_y == to.y) {
                    timeout = setTimeout(checkPosition, timeToFinishMove);
                }
                else {
                    this.socket.removeListener("player", checkPlayer);
                    reject(`move to ${to.x}, ${to.y} failed (we're currently going to ${this.going_x}, ${this.going_y})`);
                }
            };
            let timeout = setTimeout(checkPosition, timeToFinishMove);
            this.socket.on("player", checkPlayer);
        });
        if (this.going_x !== x || this.going_y !== y) {
            this.socket.emit("move", {
                going_x: to.x,
                going_y: to.y,
                m: this.m,
                x: this.x,
                y: this.y,
            });
            this.updatePositions();
            this.going_x = to.x;
            this.going_y = to.y;
            this.moving = true;
            this.move_num += 1;
        }
        return moveFinished;
    }
    openChest(id) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [openChest].");
        const chestOpened = new Promise((resolve, reject) => {
            const openCheck = (data) => {
                if (data.id == id) {
                    this.socket.removeListener("chest_opened", openCheck);
                    resolve(data);
                }
            };
            setTimeout(() => {
                this.socket.removeListener("chest_opened", openCheck);
                reject(`openChest timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("chest_opened", openCheck);
        });
        this.socket.emit("open_chest", { id: id });
        return chestOpened;
    }
    regenHP() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [regenHP].");
        const regenReceived = new Promise((resolve, reject) => {
            const regenCheck = (data) => {
                if (data.code && data.code.includes("pot_timeout")) {
                    this.socket.removeListener("eval", regenCheck);
                    resolve();
                }
            };
            setTimeout(() => {
                this.socket.removeListener("eval", regenCheck);
                reject(`regenHP timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("eval", regenCheck);
        });
        this.socket.emit("use", { item: "hp" });
        return regenReceived;
    }
    regenMP() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [regenMP].");
        const regenReceived = new Promise((resolve, reject) => {
            const regenCheck = (data) => {
                if (data.code && data.code.includes("pot_timeout")) {
                    this.socket.removeListener("eval", regenCheck);
                    resolve();
                }
            };
            setTimeout(() => {
                this.socket.removeListener("eval", regenCheck);
                reject(`regenMP timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("eval", regenCheck);
        });
        this.socket.emit("use", { item: "mp" });
        return regenReceived;
    }
    respawn() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [respawn].");
        const respawned = new Promise((resolve, reject) => {
            const respawnCheck = (data) => {
                if (data.effect == 1) {
                    this.socket.removeListener("new_map", respawnCheck);
                    this.socket.removeListener("game_log", failCheck);
                    resolve({ map: data.name, x: data.x, y: data.y });
                }
            };
            const failCheck = (data) => {
                if (data == "Can't respawn yet.") {
                    this.socket.removeListener("new_map", respawnCheck);
                    this.socket.removeListener("game_log", failCheck);
                    reject(data);
                }
            };
            setTimeout(() => {
                this.socket.removeListener("new_map", respawnCheck);
                this.socket.removeListener("game_log", failCheck);
                reject(`respawn timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("new_map", respawnCheck);
            this.socket.on("game_log", failCheck);
        });
        this.socket.emit("respawn");
        return respawned;
    }
    scare() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [scare].");
        const scared = new Promise((resolve, reject) => {
            let ids;
            const idsCheck = (data) => {
                if (data.type == "scare") {
                    ids = data.ids;
                    this.socket.removeListener("ui", idsCheck);
                }
            };
            const cooldownCheck = (data) => {
                if (/skill_timeout\s*\(\s*['"]scare['"]\s*,?\s*(\d+\.?\d+?)?\s*\)/.test(data.code)) {
                    this.socket.removeListener("ui", idsCheck);
                    this.socket.removeListener("eval", cooldownCheck);
                    resolve(ids);
                }
            };
            setTimeout(() => {
                this.socket.removeListener("ui", idsCheck);
                this.socket.removeListener("eval", cooldownCheck);
                reject(`scare timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("ui", idsCheck);
            this.socket.on("eval", cooldownCheck);
        });
        this.socket.emit("skill", { name: "scare" });
        return scared;
    }
    async sell(itemPos, quantity = 1) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [sell].");
        this.socket.emit("sell", { num: itemPos, quantity: quantity });
    }
    sendCM(to, message) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [sendCM].");
        this.socket.emit("cm", { message: JSON.stringify(message), to: to });
    }
    async sendGold(to, amount) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [sendGold].");
        if (this.gold == 0)
            return Promise.reject("We have no gold to send.");
        if (!this.players.has(to))
            return Promise.reject(`We can't see ${to} nearby to send gold.`);
        if (Tools_1.Tools.distance(this, this.players.get(to)) > Constants_1.Constants.NPC_INTERACTION_DISTANCE)
            return Promise.reject(`We are too far away from ${to} to send gold.`);
        const goldSent = new Promise((resolve, reject) => {
            const sentCheck = (data) => {
                if (data == "trade_get_closer") {
                    this.socket.removeListener("game_response", sentCheck);
                    reject(`We are too far away from ${to} to send gold.`);
                }
                else if (typeof data == "object" && data.response == "gold_sent" && data.name == to) {
                    if (data.gold !== amount)
                        console.warn(`We wanted to send ${to} ${amount} gold, but we sent ${data.gold}.`);
                    this.socket.removeListener("game_response", sentCheck);
                    resolve(data.gold);
                }
            };
            setTimeout(() => {
                this.socket.removeListener("game_response", sentCheck);
                reject(`sendGold timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("game_response", sentCheck);
        });
        this.socket.emit("send", { gold: amount, name: to });
        return goldSent;
    }
    sendItem(to, inventoryPos, quantity = 1) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [sendItem].");
        if (!this.players.has(to))
            return Promise.reject(`${to} is not nearby.`);
        if (!this.items[inventoryPos])
            return Promise.reject(`No item in inventory slot ${inventoryPos}.`);
        if (this.items[inventoryPos]?.q < quantity)
            return Promise.reject(`We only have a quantity of ${this.items[inventoryPos].q}, not ${quantity}.`);
        const item = this.items[inventoryPos];
        const itemSent = new Promise((resolve, reject) => {
            const sentCheck = (data) => {
                if (data == "trade_get_closer") {
                    this.socket.removeListener("game_response", sentCheck);
                    reject(`sendItem failed, ${to} is too far away`);
                }
                else if (data == "send_no_space") {
                    this.socket.removeListener("game_response", sentCheck);
                    reject(`sendItem failed, ${to} has no inventory space`);
                }
                else if (typeof data == "object" && data.response == "item_sent" && data.name == to && data.item == item.name && data.q == quantity) {
                    this.socket.removeListener("game_response", sentCheck);
                    resolve();
                }
            };
            setTimeout(() => {
                this.socket.removeListener("game_response", sentCheck);
                reject(`sendItem timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("game_response", sentCheck);
        });
        this.socket.emit("send", { name: to, num: inventoryPos, q: quantity });
        return itemSent;
    }
    sendPartyInvite(id) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [sendPartyInvite].");
        const invited = new Promise((resolve, reject) => {
            const sentCheck = (data) => {
                if (data == `Invited ${id} to party`) {
                    this.socket.removeListener("game_log", sentCheck);
                    resolve();
                }
            };
            setTimeout(() => {
                this.socket.removeListener("game_log", sentCheck);
                reject(`sendPartyInvite timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("game_log", sentCheck);
        });
        this.socket.emit("party", { event: "invite", name: id });
        return invited;
    }
    sendPartyRequest(id) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [sendPartyRequest].");
        this.socket.emit("party", { event: "request", name: id });
    }
    shiftBooster(booster, to) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [shiftBooster].");
        const itemInfo = this.items[booster];
        if (!itemInfo)
            return Promise.reject(`Inventory Slot ${booster} is empty.`);
        if (!["goldbooster", "luckbooster", "xpbooster"].includes(itemInfo.name))
            return Promise.reject(`The given item is not a booster (it's a '${itemInfo.name}')`);
        this.socket.emit("booster", { num: booster, action: "shift", to: to });
    }
    async smartMove(to, options = {
        avoidTownWarps: false,
        getWithin: 0,
        useBlink: false
    }) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [smartMove].");
        let fixedTo;
        let path;
        if (typeof to == "string") {
            const gMap = this.G.maps[to];
            if (gMap) {
                const mainSpawn = this.G.maps[to].spawns[0];
                fixedTo = { map: to, x: mainSpawn[0], y: mainSpawn[1] };
            }
            if (!fixedTo) {
                const gMonster = this.G.monsters[to];
                if (gMonster) {
                    const locations = this.locateMonster(to);
                    let closestDistance = Number.MAX_VALUE;
                    for (const location of locations) {
                        const potentialPath = await Pathfinder_1.Pathfinder.getPath(this, location, options?.avoidTownWarps == true);
                        const distance = Pathfinder_1.Pathfinder.computePathCost(potentialPath);
                        if (distance < closestDistance) {
                            path = potentialPath;
                            fixedTo = path[path.length - 1];
                            closestDistance = distance;
                        }
                    }
                }
            }
            if (!fixedTo) {
                const locations = this.locateNPC(to);
                let closestDistance = Number.MAX_VALUE;
                for (const location of locations) {
                    const potentialPath = await Pathfinder_1.Pathfinder.getPath(this, location, options?.avoidTownWarps == true);
                    const distance = Pathfinder_1.Pathfinder.computePathCost(potentialPath);
                    if (distance < closestDistance) {
                        path = potentialPath;
                        fixedTo = path[path.length - 1];
                        closestDistance = distance;
                    }
                }
            }
            if (!fixedTo) {
                const gItem = this.G.items[to];
                if (gItem) {
                    for (const map in this.G.maps) {
                        if (this.G.maps[map].ignore)
                            continue;
                        for (const npc of this.G.maps[map].npcs) {
                            if (this.G.npcs[npc.id].items === undefined)
                                continue;
                            for (const i of this.G.npcs[npc.id].items) {
                                if (i == to) {
                                    return this.smartMove(npc, options);
                                }
                            }
                        }
                    }
                }
            }
            if (!fixedTo)
                return Promise.reject(`Could not find a suitable destination for '${to}'`);
        }
        else if (to.x !== undefined && to.y !== undefined) {
            if (to.map)
                fixedTo = to;
            else
                fixedTo = { map: this.map, x: to.x, y: to.y };
        }
        else {
            console.debug(to);
            return Promise.reject("'to' is unsuitable for smartMove. We need a 'map', an 'x', and a 'y'.");
        }
        if (options?.getWithin >= Tools_1.Tools.distance(this, fixedTo))
            return Promise.resolve({ x: this.x, y: this.y, map: this.map });
        if (!path)
            path = await Pathfinder_1.Pathfinder.getPath(this, fixedTo, options?.avoidTownWarps == true);
        const started = Date.now();
        this.lastSmartMove = started;
        let lastMove = -1;
        for (let i = 0; i < path.length; i++) {
            let currentMove = path[i];
            if (started < this.lastSmartMove) {
                if (typeof to == "string")
                    return Promise.reject(`smartMove to ${to} cancelled (new smartMove started)`);
                else
                    return Promise.reject(`smartMove to ${to.map}:${to.x},${to.y} cancelled (new smartMove started)`);
            }
            if (options?.getWithin >= Tools_1.Tools.distance(this, fixedTo)) {
                break;
            }
            if (currentMove.type == "move" && this.map == fixedTo.map && options?.getWithin > 0) {
                const angle = Math.atan2(this.y - fixedTo.y, this.x - fixedTo.x);
                const potentialMove = {
                    type: "move",
                    map: this.map,
                    x: fixedTo.x + Math.cos(angle) * options.getWithin,
                    y: fixedTo.y + Math.sin(angle) * options.getWithin
                };
                if (Pathfinder_1.Pathfinder.canWalkPath(this, potentialMove)) {
                    i = path.length;
                    currentMove = potentialMove;
                }
            }
            if (currentMove.type == "move") {
                for (let j = i + 1; j < path.length; j++) {
                    const potentialMove = path[j];
                    if (potentialMove.map !== currentMove.map)
                        break;
                    if (potentialMove.type == "town")
                        break;
                    if (potentialMove.type == "move" && Pathfinder_1.Pathfinder.canWalkPath(this, potentialMove)) {
                        i = j;
                        currentMove = potentialMove;
                    }
                }
            }
            if (options?.useBlink && this.canUse("blink")) {
                let blinked = false;
                for (let j = path.length - 1; j > i; j--) {
                    const potentialMove = path[j];
                    if (potentialMove.map == currentMove.map) {
                        await this.blink(potentialMove.x, potentialMove.y);
                        i = j;
                        blinked = true;
                        break;
                    }
                }
                if (blinked)
                    continue;
            }
            try {
                if (currentMove.type == "leave") {
                    await this.leaveMap();
                }
                else if (currentMove.type == "move") {
                    if (currentMove.map !== this.map) {
                        return Promise.reject(`We are supposed to be in ${currentMove.map}, but we are in ${this.map}`);
                    }
                    await this.move(currentMove.x, currentMove.y, { disableSafetyCheck: true });
                }
                else if (currentMove.type == "town") {
                    await this.warpToTown();
                }
                else if (currentMove.type == "transport") {
                    await this.transport(currentMove.map, currentMove.spawn);
                }
            }
            catch (e) {
                console.error(e);
                await this.requestPlayerData();
                if (lastMove == i)
                    return Promise.reject("We are having some trouble smartMoving...");
                lastMove = i;
                i--;
            }
        }
        return { map: this.map, x: this.x, y: this.y };
    }
    async stopSmartMove() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [stopSmartMove].");
        this.lastSmartMove = Date.now();
        if (this?.c?.town) {
            this.stopWarpToTown();
        }
        return this.move(this.x, this.y);
    }
    stopWarpToTown() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [stopWarpToTown].");
        this.socket.emit("stop", { action: "town" });
    }
    takeMailItem(mailID) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [takeMailItem].");
        const itemReceived = new Promise((resolve, reject) => {
            const successCheck = (data) => {
                if (typeof data == "object") {
                    if (data.response == "mail_item_taken") {
                        this.socket.removeListener("game_response", successCheck);
                        resolve();
                    }
                }
            };
            setTimeout(() => {
                this.socket.removeListener("game_response", successCheck);
                reject("takeMailItem timeout (5000ms)");
            }, 5000);
            this.socket.on("game_response", successCheck);
        });
        this.socket.emit("mail_take_item", { id: mailID });
        return itemReceived;
    }
    throwSnowball(target, snowball = this.locateItem("snowball")) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [throwSnowball].");
        if (this.G.skills.snowball.mp > this.mp)
            return Promise.reject("Not enough MP to throw a snowball.");
        if (snowball === undefined)
            return Promise.reject("We don't have any snowballs in our inventory.");
        const throwStarted = new Promise((resolve, reject) => {
            let projectile;
            const attackCheck = (data) => {
                if (data.attacker == this.id
                    && data.type == "snowball"
                    && data.target == target) {
                    projectile = data.pid;
                }
            };
            const cooldownCheck = (data) => {
                if (/skill_timeout\s*\(\s*['"]snowball['"]\s*,?\s*(\d+\.?\d+?)?\s*\)/.test(data.code)) {
                    this.socket.removeListener("action", attackCheck);
                    this.socket.removeListener("eval", cooldownCheck);
                    resolve(projectile);
                }
            };
            const failCheck = (data) => {
                if (typeof data == "object") {
                    if (data.response == "attack_failed") {
                        this.socket.removeListener("action", attackCheck);
                        this.socket.removeListener("eval", cooldownCheck);
                        reject("throwsnowball failed");
                    }
                }
            };
            setTimeout(() => {
                this.socket.removeListener("action", attackCheck);
                this.socket.removeListener("eval", cooldownCheck);
                reject(`throwsnowball timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("action", attackCheck);
            this.socket.on("eval", cooldownCheck);
            this.socket.on("game_response", failCheck);
        });
        this.socket.emit("skill", { name: "snowball", id: target, num: snowball });
        return throwStarted;
    }
    transport(map, spawn) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [transport].");
        const transportComplete = new Promise((resolve, reject) => {
            const transportCheck = (data) => {
                this.socket.removeListener("game_response", failCheck);
                if (data.name == map)
                    resolve();
                else
                    reject(`We are now in ${data.name}, but we should be in ${map}`);
            };
            const failCheck = (data) => {
                if (typeof data == "object") {
                    if (data.response == "bank_opx" && data.reason == "mounted") {
                        this.socket.removeListener("game_response", failCheck);
                        this.socket.removeListener("new_map", transportCheck);
                        reject(`${data.name} is currently in the bank, we can't enter.`);
                    }
                }
            };
            setTimeout(() => {
                this.socket.removeListener("game_response", failCheck);
                this.socket.removeListener("new_map", transportCheck);
                reject(`transport timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.once("new_map", transportCheck);
            this.socket.on("game_response", failCheck);
        });
        this.socket.emit("transport", { to: map, s: spawn });
        return transportComplete;
    }
    unequip(slot) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [unequip].");
        if (this.slots[slot] === null)
            return Promise.reject(`Slot ${slot} is empty; nothing to unequip.`);
        if (this.slots[slot] === undefined)
            return Promise.reject(`Slot ${slot} does not exist.`);
        const unequipped = new Promise((resolve, reject) => {
            const unequipCheck = (data) => {
                if (data.slots[slot] === null) {
                    this.socket.removeListener("player", unequipCheck);
                    resolve();
                }
            };
            setTimeout(() => {
                this.socket.removeListener("player", unequipCheck);
                reject(`unequip timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("player", unequipCheck);
        });
        this.socket.emit("unequip", { slot: slot });
        return unequipped;
    }
    upgrade(itemPos, scrollPos, offeringPos) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [upgrade].");
        if (this.G.maps[this.map].mount)
            return Promise.reject("We can't upgrade things in the bank.");
        const itemInfo = this.items[itemPos];
        const scrollInfo = this.items[scrollPos];
        if (!itemInfo)
            return Promise.reject(`There is no item in inventory slot ${itemPos}.`);
        if (this.G.items[itemInfo.name].upgrade == undefined)
            return Promise.reject("This item is not upgradable.");
        if (!scrollInfo)
            return Promise.reject(`There is no scroll in inventory slot ${scrollPos}.`);
        if (offeringPos !== undefined) {
            const offeringInfo = this.items[offeringPos];
            if (!offeringInfo)
                return Promise.reject(`There is no item in inventory slot ${offeringPos} (offering).`);
        }
        const upgradeComplete = new Promise((resolve, reject) => {
            const playerCheck = (data) => {
                if (!data.hitchhikers)
                    return;
                for (const [event, datum] of data.hitchhikers) {
                    if (event == "game_response" && datum.response == "upgrade_fail" && datum.num == itemPos) {
                        this.socket.removeListener("game_response", gameResponseCheck);
                        this.socket.removeListener("player", playerCheck);
                        resolve(false);
                        return;
                    }
                    else if (event == "game_response" && datum.response == "upgrade_success" && datum.num == itemPos) {
                        this.socket.removeListener("game_response", gameResponseCheck);
                        this.socket.removeListener("player", playerCheck);
                        resolve(true);
                        return;
                    }
                }
            };
            const gameResponseCheck = (data) => {
                if (typeof data == "object" && data.response == "bank_restrictions" && data.place == "upgrade") {
                    this.socket.removeListener("game_response", gameResponseCheck);
                    this.socket.removeListener("player", playerCheck);
                    reject("You can't upgrade items in the bank.");
                }
                else if (typeof data == "string") {
                    if (data == "bank_restrictions") {
                        this.socket.removeListener("game_response", gameResponseCheck);
                        this.socket.removeListener("player", playerCheck);
                        reject("We can't upgrade things in the bank.");
                    }
                    else if (data == "upgrade_in_progress") {
                        this.socket.removeListener("game_response", gameResponseCheck);
                        this.socket.removeListener("player", playerCheck);
                        reject("We are already upgrading something.");
                    }
                    else if (data == "upgrade_incompatible_scroll") {
                        this.socket.removeListener("game_response", gameResponseCheck);
                        this.socket.removeListener("player", playerCheck);
                        reject(`The scroll we're trying to use (${scrollInfo.name}) isn't a high enough grade to upgrade this item.`);
                    }
                    else if (data == "upgrade_success") {
                        this.socket.removeListener("game_response", gameResponseCheck);
                        this.socket.removeListener("player", playerCheck);
                        resolve(true);
                    }
                    else if (data == "upgrade_fail") {
                        this.socket.removeListener("game_response", gameResponseCheck);
                        this.socket.removeListener("player", playerCheck);
                        resolve(false);
                    }
                }
            };
            setTimeout(() => {
                this.socket.removeListener("game_response", gameResponseCheck);
                this.socket.removeListener("player", playerCheck);
                reject("upgrade timeout (60000ms)");
            }, 60000);
            this.socket.on("game_response", gameResponseCheck);
            this.socket.on("player", playerCheck);
        });
        this.socket.emit("upgrade", { item_num: itemPos, scroll_num: scrollPos, offering_num: offeringPos, clevel: this.items[itemPos].level });
        return upgradeComplete;
    }
    useHPPot(itemPos) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [useHPPot].");
        if (!this.items[itemPos])
            return Promise.reject(`There is no item in inventory slot ${itemPos}.`);
        if (this.G.items[this.items[itemPos].name].type !== "pot")
            return Promise.reject(`The item provided (${itemPos}) is not a potion.`);
        if (this.G.items[this.items[itemPos].name].gives[0][0] !== "hp")
            return Promise.reject(`The item provided(${itemPos}) is not an HP Potion.`);
        if (this.G.items[this.items[itemPos].name].gives[0][1] < 0)
            return Promise.reject(`The item provided(${itemPos}) is not an HP Potion.`);
        const healReceived = new Promise((resolve, reject) => {
            const healCheck = (data) => {
                if (data.code && data.code.includes("pot_timeout")) {
                    this.socket.removeListener("eval", healCheck);
                    resolve();
                }
            };
            setTimeout(() => {
                this.socket.removeListener("eval", healCheck);
                reject(`useHPPot timeout(${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("eval", healCheck);
        });
        this.socket.emit("equip", { num: itemPos });
        return healReceived;
    }
    useMPPot(itemPos) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [useMPPot].");
        if (!this.items[itemPos])
            return Promise.reject(`There is no item in inventory slot ${itemPos}.`);
        if (this.G.items[this.items[itemPos].name].type !== "pot")
            return Promise.reject(`The item provided (${itemPos}) is not a potion.`);
        if (this.G.items[this.items[itemPos].name].gives[0][0] !== "mp")
            return Promise.reject(`The item provided(${itemPos}) is not an MP Potion.`);
        if (this.G.items[this.items[itemPos].name].gives[0][1] < 0)
            return Promise.reject(`The item provided(${itemPos}) is not an MP Potion.`);
        const healReceived = new Promise((resolve, reject) => {
            const healCheck = (data) => {
                if (data.code && data.code.includes("pot_timeout")) {
                    this.socket.removeListener("eval", healCheck);
                    resolve();
                }
            };
            setTimeout(() => {
                this.socket.removeListener("eval", healCheck);
                reject(`useMPPot timeout (${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("eval", healCheck);
        });
        this.socket.emit("equip", { num: itemPos });
        return healReceived;
    }
    warpToJail() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [warpToJail].");
        return this.move(Number.MAX_VALUE, Number.MAX_VALUE, { disableSafetyCheck: true });
    }
    warpToTown() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [warpToTown].");
        let startedWarp = false;
        const warpComplete = new Promise((resolve, reject) => {
            const failCheck = (data) => {
                if (!startedWarp && data.c.town && data.c.town.ms == 3000) {
                    startedWarp = true;
                    return;
                }
                if (startedWarp && !data.c.town) {
                    this.socket.removeListener("player", failCheck);
                    this.socket.removeListener("new_map", warpedCheck2);
                    reject("warpToTown failed.");
                }
            };
            const warpedCheck2 = (data) => {
                if (data.effect == 1) {
                    this.socket.removeListener("player", failCheck);
                    this.socket.removeListener("new_map", warpedCheck2);
                    resolve({ map: data.name, x: data.x, y: data.y });
                }
            };
            setTimeout(() => {
                if (!startedWarp) {
                    this.socket.removeListener("player", failCheck);
                    this.socket.removeListener("new_map", warpedCheck2);
                    reject("warpToTown timeout (1000ms)");
                }
            }, 1000);
            setTimeout(() => {
                this.socket.removeListener("player", failCheck);
                this.socket.removeListener("new_map", warpedCheck2);
                reject("warpToTown timeout (5000ms)");
            }, 5000);
            this.socket.on("new_map", warpedCheck2);
            this.socket.on("player", failCheck);
        });
        this.socket.emit("town");
        return warpComplete;
    }
    withdrawGold(gold) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [withdrawGold].");
        if (this.map !== "bank")
            return Promise.reject("We need to be in 'bank' to withdraw gold.");
        if (gold <= 0)
            return Promise.reject("We can't withdraw 0 or less gold.");
        if (gold > this.bank.gold) {
            gold = this.bank.gold;
            console.warn(`We are only going to withdraw ${gold} gold.`);
        }
        this.socket.emit("bank", { operation: "withdraw", amount: gold });
    }
    withdrawItem(bankPack, bankPos, inventoryPos = -1) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [withdrawItem].");
        const item = this.bank[bankPack][bankPos];
        if (!item)
            return Promise.reject(`There is no item in bank ${bankPack}[${bankPos}]`);
        const bankPackNum = Number.parseInt(bankPack.substr(5, 2));
        if ((this.map == "bank" && bankPackNum < 0 && bankPackNum > 7)
            || (this.map == "bank_b" && bankPackNum < 8 && bankPackNum > 23)
            || (this.map == "bank_u" && bankPackNum < 24 && bankPackNum > 47)) {
            return Promise.reject(`We can not access ${bankPack} on ${this.map}.`);
        }
        const itemCount = this.countItem(item.name);
        const swapped = new Promise((resolve, reject) => {
            const checkWithdrawal = (data) => {
                const newCount = this.countItem(item.name, data.items);
                if (item.q && newCount == (itemCount + item.q)
                    || !item.q && newCount == (itemCount + 1)) {
                    this.socket.removeListener("player", checkWithdrawal);
                    return resolve();
                }
            };
            setTimeout(() => {
                this.socket.removeListener("player", checkWithdrawal);
                reject(`withdrawItem timeout(${Constants_1.Constants.TIMEOUT}ms)`);
            }, Constants_1.Constants.TIMEOUT);
            this.socket.on("player", checkWithdrawal);
        });
        this.socket.emit("bank", { operation: "swap", pack: bankPack, str: bankPos, inv: inventoryPos });
        return swapped;
    }
    countItem(item, inventory = this.items, filters) {
        let count = 0;
        for (const inventoryItem of inventory) {
            if (!inventoryItem)
                continue;
            if (inventoryItem.name !== item)
                continue;
            if (filters?.levelGreaterThan !== undefined) {
                if (inventoryItem.level == undefined)
                    continue;
                if (inventoryItem.level <= filters.levelGreaterThan)
                    continue;
            }
            if (filters?.levelLessThan !== undefined) {
                if (inventoryItem.level == undefined)
                    continue;
                if (inventoryItem.level >= filters.levelLessThan)
                    continue;
            }
            if (inventoryItem.q) {
                count += inventoryItem.q;
            }
            else {
                count += 1;
            }
        }
        return count;
    }
    getCooldown(skill) {
        if (this.G.skills[skill].share)
            skill = this.G.skills[skill].share;
        const nextSkill = this.nextSkill.get(skill);
        if (!nextSkill)
            return 0;
        const cooldown = nextSkill.getTime() - Date.now();
        if (cooldown < 0)
            return 0;
        return cooldown;
    }
    getNearestAttackablePlayer() {
        if (!this.isPVP())
            return undefined;
        let closest;
        let closestD = Number.MAX_VALUE;
        for (const [, player] of this.players) {
            if (player.s?.invincible)
                return;
            if (player.npc)
                return;
            const d = Tools_1.Tools.distance(this, player);
            if (d < closestD) {
                closest = player;
                closestD = d;
            }
        }
        if (closest)
            return { player: closest, distance: closestD };
    }
    hasPvPMarkedItem(inv = this.items) {
        for (let i = 0; i < inv.length; i++) {
            const item = inv[i];
            if (!item)
                continue;
            if (item.v)
                return true;
        }
        return false;
    }
    hasItem(iN, inv = this.items, args) {
        return this.locateItem(iN, inv, args) !== undefined;
    }
    isEquipped(itemName) {
        for (const slot in this.slots) {
            if (!this.slots[slot])
                continue;
            if (this.slots[slot].b)
                continue;
            if (this.slots[slot].name == itemName)
                return true;
        }
        return false;
    }
    isOnCooldown(skill) {
        return this.getCooldown(skill) !== 0;
    }
    isPVP() {
        if (this.G.maps[this.map].pvp)
            return true;
        if (this.G.maps[this.map].safe)
            return false;
        return this.server.pvp;
    }
    locateDuplicateItems(inventory = this.items) {
        const items = [];
        for (let i = 0; i < inventory.length; i++) {
            const item = inventory[i];
            if (!item)
                continue;
            items.push({ ...item, slotNum: i });
        }
        items.sort((a, b) => {
            const n = a.name.localeCompare(b.name);
            if (n !== 0)
                return n;
            if (a.level !== undefined && b.level !== undefined && a.level !== b.level)
                return a.level - b.level;
            if (a.p !== undefined && b.p === undefined)
                return -1;
            return 0;
        });
        const duplicates = {};
        for (let i = 0; i < items.length - 1; i++) {
            const item1 = items[i];
            for (let j = i + 1; j < items.length; j++) {
                const item2 = items[j];
                if (item1.name === item2.name) {
                    if (j == i + 1) {
                        duplicates[item1.name] = [item1.slotNum, item2.slotNum];
                    }
                    else {
                        duplicates[item1.name].push(item2.slotNum);
                    }
                }
                else {
                    i = j - 1;
                    break;
                }
            }
        }
        return duplicates;
    }
    locateItem(itemName, inventory = this.items, filters) {
        if (filters?.quantityGreaterThan == 0)
            delete filters.quantityGreaterThan;
        for (let i = 0; i < inventory.length; i++) {
            const item = inventory[i];
            if (!item)
                continue;
            if (item.name !== itemName)
                continue;
            if (filters?.level !== undefined) {
                if (item.level !== filters.level)
                    continue;
            }
            if (filters?.levelGreaterThan !== undefined) {
                if (item.level == undefined)
                    continue;
                if (item.level <= filters.levelGreaterThan)
                    continue;
            }
            if (filters?.levelLessThan !== undefined) {
                if (item.level == undefined)
                    continue;
                if (item.level >= filters.levelLessThan)
                    continue;
            }
            if (filters?.quantityGreaterThan !== undefined) {
                if (item.q == undefined)
                    continue;
                if (item.q <= filters.quantityGreaterThan)
                    continue;
            }
            return i;
        }
        return undefined;
    }
    locateItems(itemName, inventory = this.items, filters) {
        const found = [];
        for (let i = 0; i < inventory.length; i++) {
            const item = inventory[i];
            if (!item)
                continue;
            if (item.name !== itemName)
                continue;
            if (filters) {
                if (filters.level !== undefined) {
                    if (item.level !== filters.level)
                        continue;
                }
                if (filters.levelGreaterThan !== undefined) {
                    if (item.level == undefined)
                        continue;
                    if (item.level <= filters.levelGreaterThan)
                        continue;
                }
                if (filters.levelLessThan !== undefined) {
                    if (item.level == undefined)
                        continue;
                    if (item.level >= filters.levelLessThan)
                        continue;
                }
                if (filters.quantityGreaterThan !== undefined) {
                    if (item.q == undefined)
                        continue;
                    if (item.q <= filters.quantityGreaterThan)
                        continue;
                }
            }
            found.push(i);
        }
        return found;
    }
    locateMonster(mType) {
        const locations = [];
        if (mType == "goldenbat")
            mType = "bat";
        else if (mType == "snowman")
            mType = "arcticbee";
        for (const mapName in this.G.maps) {
            if (this.G.maps[mapName].ignore)
                continue;
            const map = this.G.maps[mapName];
            if (map.instance || !map.monsters || map.monsters.length == 0)
                continue;
            for (const monsterSpawn of map.monsters) {
                if (monsterSpawn.type != mType)
                    continue;
                if (monsterSpawn.boundary) {
                    locations.push({ "map": mapName, "x": (monsterSpawn.boundary[0] + monsterSpawn.boundary[2]) / 2, "y": (monsterSpawn.boundary[1] + monsterSpawn.boundary[3]) / 2 });
                }
                else if (monsterSpawn.boundaries) {
                    for (const boundary of monsterSpawn.boundaries) {
                        locations.push({ "map": boundary[0], "x": (boundary[1] + boundary[3]) / 2, "y": (boundary[2] + boundary[4]) / 2 });
                    }
                }
            }
        }
        return locations;
    }
    locateNPC(npcID) {
        const locations = [];
        for (const mapName in this.G.maps) {
            const map = this.G.maps[mapName];
            if (map.ignore)
                continue;
            if (map.instance || !map.npcs || map.npcs.length == 0)
                continue;
            for (const npc of map.npcs) {
                if (npc.id !== npcID)
                    continue;
                if (npc.position) {
                    locations.push({ map: mapName, x: npc.position[0], y: npc.position[1] });
                }
                else if (npc.positions) {
                    for (const position of npc.positions) {
                        locations.push({ map: mapName, x: position[0], y: position[1] });
                    }
                }
            }
        }
        return locations;
    }
    locateCraftNPC(itemName) {
        const gCraft = this.G.craft[itemName];
        if (gCraft) {
            const npcToLocate = gCraft.quest ? gCraft.quest : "craftsman";
            for (const mapName in this.G.maps) {
                const gMap = this.G.maps[mapName];
                if (gMap.ignore)
                    continue;
                for (const npc of gMap.npcs) {
                    if (npc.id == npcToLocate) {
                        return { map: mapName, x: npc.position[0], y: npc.position[1] };
                    }
                }
            }
        }
        throw Error(`${itemName} is not craftable.`);
    }
    locateExchangeNPC(itemName) {
        const gItem = this.G.items[itemName];
        if (gItem.quest) {
            let npcToLocate;
            for (const npcName in this.G.npcs) {
                const gNPC = this.G.npcs[npcName];
                if (gNPC.ignore)
                    continue;
                if (gNPC.quest == gItem.quest) {
                    npcToLocate = gNPC.id;
                    break;
                }
            }
            if (npcToLocate) {
                for (const mapName in this.G.maps) {
                    const gMap = this.G.maps[mapName];
                    if (gMap.ignore)
                        continue;
                    for (const npc of gMap.npcs) {
                        if (npc.id == npcToLocate) {
                            return { map: mapName, x: npc.position[0], y: npc.position[1] };
                        }
                    }
                }
            }
        }
        if (gItem.type == "token") {
            let npcToLocate;
            for (const npcName in this.G.npcs) {
                const gNPC = this.G.npcs[npcName];
                if (gNPC.ignore)
                    continue;
                if (gNPC.token == itemName) {
                    npcToLocate = gNPC.id;
                    break;
                }
            }
            if (npcToLocate) {
                for (const mapName in this.G.maps) {
                    const gMap = this.G.maps[mapName];
                    if (gMap.ignore)
                        continue;
                    for (const npc of gMap.npcs) {
                        if (npc.id == npcToLocate) {
                            return { map: mapName, x: npc.position[0], y: npc.position[1] };
                        }
                    }
                }
            }
        }
        if (gItem.e) {
            for (const mapName in this.G.maps) {
                const gMap = this.G.maps[mapName];
                if (gMap.ignore)
                    continue;
                for (const npc of gMap.npcs) {
                    if (npc.id == "exchange") {
                        return { map: mapName, x: npc.position[0], y: npc.position[1] };
                    }
                }
            }
        }
        throw Error(`${itemName} is not exchangable`);
    }
}
exports.Character = Character;
//# sourceMappingURL=Character.js.map